"""
This module handles server configuration.

See profiles.py for client configuration.
"""

import copy
import os
import warnings
from collections import defaultdict
from datetime import timedelta
from functools import cache
from pathlib import Path
from typing import Any, Union

import jsonschema

from .adapters.mapping import MapAdapter
from .catalog import from_uri, in_memory
from .media_type_registration import (
    default_compression_registry,
    default_deserialization_registry,
    default_serialization_registry,
)
from .query_registration import default_query_registry
from .server.settings import Settings
from .structures.core import Spec
from .utils import import_object, parse, prepend_to_sys_path
from .validation_registration import default_validation_registry


@cache
def schema():
    "Load the schema for service-side configuration."
    import yaml

    here = Path(__file__).parent.absolute()
    schema_path = here / "config_schemas" / "service_configuration.yml"
    with open(schema_path, "r") as file:
        return yaml.safe_load(file)


def construct_build_app_kwargs(
    config, *, source_filepath: Union[Path, str, None] = None
):
    """
    Given parsed configuration, construct arguments for build_app(...).

    The singleton global instances of the registries are used (and modified).
    """
    config = copy.deepcopy(config)  # Avoid mutating input.
    startup_tasks = []
    shutdown_tasks = []
    background_tasks = []

    sys_path_additions = []
    if source_filepath:
        if os.path.isdir(source_filepath):
            directory = source_filepath
        else:
            directory = os.path.dirname(source_filepath)
        sys_path_additions.append(directory)
    with prepend_to_sys_path(*sys_path_additions):
        # Process auth settings
        auth_spec = config.get("authentication", {}) or {}
        for age in ["refresh_token_max_age", "session_max_age", "access_token_max_age"]:
            if age in auth_spec:
                auth_spec[age] = timedelta(seconds=auth_spec[age])
        access_control = config.get("access_control", {}) or {}
        providers = list(auth_spec.get("providers", []))
        provider_names = [p["provider"] for p in providers]
        if len(set(provider_names)) != len(provider_names):
            raise ValueError(
                "The names given for 'provider' must be unique. "
                f"Found duplicates in {providers}"
            )
        for i, authenticator in enumerate(providers):
            authenticator_class = import_object(
                authenticator["authenticator"], accept_live_object=True
            )
            authenticator = authenticator_class(**authenticator.get("args", {}))
            # Replace "package.module:Object" with live instance.
            auth_spec["providers"][i]["authenticator"] = authenticator
        if access_control.get("access_policy") is not None:
            access_policy_import_path = access_control["access_policy"]
            access_policy_class = import_object(
                access_policy_import_path, accept_live_object=True
            )
            access_policy = access_policy_class(**access_control.get("args", {}))
        else:
            access_policy = None
        # TODO Enable entrypoint to extend aliases?

        # Process server settings
        server_settings = {}
        if root_path := config.get("root_path", ""):
            server_settings["root_path"] = root_path
        server_settings["allow_origins"] = config.get("allow_origins")
        server_settings["response_bytesize_limit"] = config.get(
            "response_bytesize_limit"
        )
        server_settings["exact_count_limit"] = config.get("exact_count_limit")
        server_settings["database"] = config.get("database", {})
        server_settings["reject_undeclared_specs"] = config.get(
            "reject_undeclared_specs"
        )
        server_settings["expose_raw_assets"] = config.get("expose_raw_assets")
        server_settings["metrics"] = config.get("metrics", {})

        # Process trees
        tree_aliases = {
            "catalog": "tiled.catalog:from_uri",
        }
        trees = {}
        for item in config.get("trees", []):
            segments = tuple(segment for segment in item["path"].split("/") if segment)
            tree_spec = item["tree"]
            if isinstance(tree_spec, str) and tree_spec == "files":
                raise Exception(
                    """The way that tiled serves files has changed.

See documentation section "Serve a Directory of Files"."""
                )
            import_path = tree_aliases.get(tree_spec, tree_spec)
            obj = import_object(import_path, accept_live_object=True)
            if ("args" in item) and (not callable(obj)):
                raise ValueError(
                    f"Object imported from {import_path} cannot take args. "
                    "It is not callable."
                )
            if callable(obj):
                # Interpret obj as a tree *factory*.
                args = {}
                args.update(item.get("args", {}))

                # Add other server-related settings falling back to Settings defaults
                # TODO: To be refactroed; these parameters should be in `server_settings`
                if obj is from_uri:
                    default_settings = Settings().model_dump()
                    from_server_settings = {
                        k: config.get(k, default_settings[k])
                        for k in {
                            "catalog_pool_size",
                            "storage_pool_size",
                            "catalog_max_overflow",
                            "storage_max_overflow",
                        }
                    }
                    args.update(from_server_settings)
                if (obj is from_uri) or (obj is in_memory):
                    args.update({"cache_settings": config.get("streaming_cache")})
                tree = obj(**args)
            else:
                # Interpret obj as a tree *instance*.
                tree = obj
            if segments in trees:
                raise ValueError(f"The path {'/'.join(segments)} was specified twice.")
            trees[segments] = tree
            startup_tasks.extend(getattr(tree, "startup_tasks", []))
            shutdown_tasks.extend(getattr(tree, "shutdown_tasks", []))
            background_tasks.extend(getattr(tree, "background_tasks", []))
        if not len(trees):
            raise ValueError("Configuration contains no trees")
        if list(trees) == [()]:
            # Simple case: there is one tree, served at the root path /.
            root_tree = tree
        else:
            # There are one or more tree(s) to be served at
            # sub-paths. Merged them into one root MapAdapter.
            # Map path segments to dicts containing Adapters at that path.
            root_mapping = {}
            index = {(): root_mapping}
            include_routers = []
            for segments, tree in trees.items():
                for i in range(len(segments)):
                    if segments[:i] not in index:
                        mapping = {}
                        index[segments[:i]] = mapping
                        parent = index[segments[: i - 1]]
                        parent[segments[i - 1]] = MapAdapter(mapping)
                index[segments[:-1]][segments[-1]] = tree
                # Collect any custom routers.
                routers = getattr(tree, "include_routers", [])
                for router in routers:
                    if router not in include_routers:
                        include_routers.append(router)
            root_tree = MapAdapter(root_mapping)
            root_tree.include_routers.extend(include_routers)

        # Process other configuration items
        for structure_family, values in config.get("media_types", {}).items():
            for media_type, import_path in values.items():
                serializer = import_object(import_path, accept_live_object=True)
                default_serialization_registry.register(
                    structure_family, media_type, serializer
                )
        for ext, media_type in config.get("file_extensions", {}).items():
            default_serialization_registry.register_alias(ext, media_type)

        for item in config.get("specs", []):
            if "validator" in item:
                validator = import_object(item["validator"])
            else:
                # no-op
                validator = _no_op_validator
            default_validation_registry.register(Spec(item["spec"]), validator)

    # TODO Make compression_registry extensible via configuration.
    return {
        "tree": root_tree,
        "authentication": auth_spec,
        "server_settings": server_settings,
        "query_registry": default_query_registry,
        "serialization_registry": default_serialization_registry,
        "deserialization_registry": default_deserialization_registry,
        "compression_registry": default_compression_registry,
        "validation_registry": default_validation_registry,
        "tasks": {
            "startup": startup_tasks,
            "shutdown": shutdown_tasks,
            "background": background_tasks,
        },
        "access_policy": access_policy,
    }


def merge(configs: dict[Path, dict[str, Any]]) -> dict[str, Any]:
    merged: dict[str, Any] = {"trees": []}

    # These variables are used to produce error messages that point
    # to the relevant config file(s).
    authentication_config_source = None
    access_control_config_source = None
    uvicorn_config_source = None
    metrics_config_source = None
    database_config_source = None
    response_bytesize_limit_config_source = None
    allow_origins = []
    media_types = defaultdict(dict)
    specs = []
    reject_undeclared_specs_source = None
    streaming_cache_source = None
    file_extensions = {}
    paths = {}  # map each item's path to config file that specified it

    for filepath, config in configs.items():
        for structure_family, values in config.get("media_types", {}).items():
            media_types[structure_family].update(values)

        file_extensions.update(config.get("file_extensions", {}))
        allow_origins.extend(config.get("allow_origins", []))
        if "access_control" in config:
            if "access_control" in merged:
                raise ConfigError(
                    "access_control can only be specified in one file. "
                    f"It was found in both {access_control_config_source} and "
                    f"{filepath}"
                )
            access_control_config_source = filepath
            merged["access_control"] = config["access_control"]
        if "authentication" in config:
            if "authentication" in merged:
                raise ConfigError(
                    "authentication can only be specified in one file. "
                    f"It was found in both {authentication_config_source} and "
                    f"{filepath}"
                )
            authentication_config_source = filepath
            merged["authentication"] = config["authentication"]
        if "uvicorn" in config:
            if "uvicorn" in merged:
                raise ConfigError(
                    "uvicorn can only be specified in one file. "
                    f"It was found in both {uvicorn_config_source} and "
                    f"{filepath}"
                )
            uvicorn_config_source = filepath
            merged["uvicorn"] = config["uvicorn"]
        if "object_cache" in config:
            warnings.warn(
                "The object cache has been removed. "
                "The config of the object cache no longer has any effect."
            )
        if "response_bytesize_limit" in config:
            if "response_bytesize_limit" in merged:
                raise ConfigError(
                    "response_bytesize_limit can only be specified in one file. "
                    f"It was found in both {response_bytesize_limit_config_source} and "
                    f"{filepath}"
                )
            response_bytesize_limit_config_source = filepath
            merged["response_bytesize_limit"] = config["response_bytesize_limit"]
        if "metrics" in config:
            if "metrics" in merged:
                raise ConfigError(
                    "metrics can only be specified in one file. "
                    f"It was found in both {metrics_config_source} and "
                    f"{filepath}"
                )
            metrics_config_source = filepath
            merged["metrics"] = config["metrics"]
        if "database" in config:
            if "database" in merged:
                raise ConfigError(
                    "database configuration can only be specified in one file. "
                    f"It was found in both {database_config_source} and "
                    f"{filepath}"
                )
            database_config_source = filepath
            merged["database"] = config["database"]
        if "reject_undeclared_specs" in config:
            if "reject_undeclared_specs" in merged:
                raise ConfigError(
                    "'reject_undeclared_specs' can only be specified in one file. "
                    f"It was found in both {reject_undeclared_specs_source} and "
                    f"{filepath}"
                )
            reject_undeclared_specs_source = filepath
            merged["reject_undeclared_specs"] = config["reject_undeclared_specs"]
        if "streaming_cache" in config:
            if "streaming_cache" in merged:
                raise ConfigError(
                    "'streaming_cache' can only be specified in one file. "
                    f"It was found in both {streaming_cache_source} and "
                    f"{filepath}"
                )
            streaming_cache_source = filepath
            merged["streaming_cache"] = config["streaming_cache"]
        for item in config.get("trees", []):
            if item["path"] in paths:
                msg = "A given path may be only be specified once."
                "The path {item['path']} was found twice in "
                if filepath == paths[item["path"]]:
                    msg += f"{filepath}."
                else:
                    msg += f"{filepath} and {paths[item['path']]}."
                raise ConfigError(msg)
            paths[item["path"]] = filepath
            merged["trees"].append(item)
        specs.extend(config.get("specs", []))
    merged["media_types"] = dict(media_types)  # convert from defaultdict
    merged["file_extensions"] = file_extensions
    merged["allow_origins"] = allow_origins
    merged["specs"] = specs
    return merged


def parse_configs(config_path: Union[str, Path]) -> dict[str, Any]:
    """
    Parse configuration file or directory of configuration files.

    If a directory is given, any files not ending in `.yml` or `.yaml` are
    ignored. Therefore, the directory may also contain a README file and
    supporting Python scripts with custom objects.
    """
    if isinstance(config_path, str):
        config_path = Path(config_path)
    if config_path.is_file():
        filepaths = [config_path]
    elif config_path.is_dir():
        filepaths = [
            fn
            for fn in config_path.iterdir()
            if fn.suffix in (".yml", ".yaml") and fn.is_file()
        ]
    elif not config_path.exists():
        raise ValueError(f"The config path {config_path!s} doesn't exist.")
    else:
        # the path points to something we don't support, eg fifo/block_device/etc
        raise ValueError(
            f"The config path {config_path!s} exists but is not a file or directory."
        )

    parsed_configs: dict[Path, dict[str, Any]] = {}
    # The sorting here is just to make the order of the results deterministic.
    # There is *not* any sorting-based precedence applied.
    for filepath in sorted(filepaths):
        # Ignore hidden files and .py files.
        if (
            filepath.parts[-1].startswith(".")
            or filepath.suffix == ".py"
            or filepath.parts[-1] == "__pycache__"
        ):
            continue
        with open(filepath) as file:
            config = parse(file)
            try:
                jsonschema.validate(instance=config, schema=schema())
            except jsonschema.ValidationError as err:
                msg = err.args[0]
                raise ConfigError(
                    f"ValidationError while parsing configuration file {filepath}: {msg}"
                ) from err
            parsed_configs[filepath] = config

    merged_config = merge(parsed_configs)
    return merged_config


class ConfigError(ValueError):
    pass


def _no_op_validator(*args, **kwargs):
    return None



import pathlib

import pytest
from fastapi import Query
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY

from ..catalog import in_memory
from ..client import Context, from_context
from ..ndslice import NDSlice
from ..server.app import build_app


@pytest.fixture(scope="module")
def module_tmp_path(tmp_path_factory: pytest.TempdirFactory) -> pathlib.Path:
    return tmp_path_factory.mktemp("temp")


@pytest.fixture(scope="module")
def client(module_tmp_path):
    catalog = in_memory(writable_storage=str(module_tmp_path))
    app = build_app(catalog)
    with Context.from_app(app) as context:
        client = from_context(context)
        client.write_array([1], key="x")
        yield client


slice_test_data = [
    "",
    ":",
    "::",
    "...",
    ":,::",
    "0",
    "0:",
    "0::",
    ":0",
    "::0",
    "5:",
    ":10",
    "::12",
    "-1",
    "-5:",
    ":-5",
    "::-45",
    "3:5",
    "5:3",
    "123::4",
    "5::678",
    ":123:4",
    ":5:678",
    "0,1,2",
    "5:,:10,::-5",
    "1:2:3,4:5:6,7:8:9",
    "10::20,30::40,50::60",
    "1 : 2",
    "1:2, 3",
    "1 ,2:3",
    "1 , 2 , 3",
]

slice_missing_data = [
    ",",
    ",,",
    ",:",
    ",,:,::,,::,:,,::,",
    ",1:2,,:3:4",
    ",:1, 1::, ,, 1:5:2" "-2:4:1,,:5:2",
]

slice_typo_data = [
    ":::",
    "..",
    "....",
    "1:2:3:4",
    "1:2,3:4:5:6",
]

slice_malicious_data = [
    "1:(2+3)",
    "1**2",
    "print('oh so innocent')",
    "; print('oh so innocent')",
    ")\"; print('oh so innocent')",
    "1:2)\"; print('oh so innocent')",
    "1:2)\";print('oh_so_innocent')",
    "import sys; sys.exit()",
    "; import sys; sys.exit()",
    "touch /tmp/x",
    "rm -rf /tmp/*",
]


# this is the outgoing slice_ function from tiled.server.dependencies as is
def reference_slice_(
    slice: str = Query(None, pattern="^[-0-9,:]*$"),
):
    "Specify and parse a block index parameter."
    import numpy

    # IMPORTANT We are eval-ing a user-provider string here so we need to be
    # very careful about locking down what can be in it. The regex above
    # excludes any letters or operators, so it is not possible to execute
    # functions or expensive arithmetic.
    return tuple(
        [
            eval(f"numpy.s_[{dim!s}]", {"numpy": numpy})
            for dim in (slice or "").split(",")
            if dim
        ]
    )


@pytest.mark.parametrize("slice", slice_test_data)
def test_slicer(slice: str):
    """
    Test the slicer function
    """
    assert NDSlice.from_query(slice) == reference_slice_(slice)


@pytest.mark.parametrize("slice", slice_typo_data + slice_missing_data)
def test_slicer_typo_data(slice: str):
    """
    Test the slicer function with invalid input
    """
    with pytest.raises(ValueError):
        _ = NDSlice.from_query(slice)


@pytest.mark.parametrize("slice", slice_malicious_data)
def test_slicer_malicious_exec(slice: str):
    """
    Test the slicer function with 'malicious' input
    """
    with pytest.raises(ValueError):
        _ = NDSlice.from_query(slice)


@pytest.mark.parametrize("slice_", slice_typo_data + slice_malicious_data)
def test_slicer_fastapi_query_rejection(slice_, client):
    http_client = client.context.http_client
    response = http_client.get(f"/api/v1/array/block/x?block=0&slice={slice_}")
    assert response.status_code == HTTP_422_UNPROCESSABLE_ENTITY


slice_cases = [
    ("[0:5:2]", (slice(0, 5, 2),), [{"start": 0, "stop": 5, "step": 2}]),
    ("(1:3)", (slice(1, 3),), [{"start": 1, "stop": 3}]),
    ("1:", (slice(1, None),), [{"start": 1}]),
    ("[0:3]", (slice(0, 3),), [{"start": 0, "stop": 3}]),
    ("[:]", (slice(None, None, None),), [{}]),
    ("[...]", (Ellipsis,), [{}]),
    ("::2", (slice(None, None, 2),), [{"step": 2}]),
    ("[1::2]", (slice(1, None, 2),), [{"start": 1, "step": 2}]),
    ("[1 : : 2]", (slice(1, None, 2),), [{"start": 1, "step": 2}]),
    ("[:3:2]", (slice(None, 3, 2),), [{"stop": 3, "step": 2}]),
    ("[1:3:]", (slice(1, 3),), [{"start": 1, "stop": 3}]),
    (
        "[1:5:2, ..., :5]",
        (slice(1, 5, 2), Ellipsis, slice(None, 5)),
        [{"start": 1, "stop": 5, "step": 2}, {}, {"stop": 5}],
    ),
]


@pytest.mark.parametrize("as_string, as_tuple, as_json", slice_cases)
def test_ndslice_conversion(as_string, as_tuple, as_json):
    assert NDSlice.from_numpy_str(as_string) == as_tuple

    # JSON does not preserve Ellipsis expanding them to empty slices
    as_tuple_with_ellipsis = tuple(
        slice(None) if x is Ellipsis else x for x in as_tuple
    )
    assert NDSlice.from_json(as_json) == as_tuple_with_ellipsis

    # Normalize the string representations before comparing them
    norm_string = as_string.strip("(][)").replace(" ", "").lstrip("0")
    norm_string = "1:3" if norm_string == "1:3:" else norm_string
    assert NDSlice(*as_tuple).to_numpy_str() == norm_string


slice_json_cases = [
    ((slice(0, 5, 2),), [{"start": 0, "stop": 5, "step": 2}], None),
    ((slice(0, 5, 2),), [{"start": 0, "stop": 5, "step": 2}], 1),
    ((slice(0, 5, 2), slice(None)), [{"start": 0, "stop": 5, "step": 2}, {}], 2),
    ((slice(0, 5, 2),), [{"start": 0, "stop": 5, "step": 2}, {}, {}], 3),
    ((slice(0, 5, 2), Ellipsis), [{"start": 0, "stop": 5, "step": 2}, {}, {}], 3),
    ((Ellipsis, slice(0, 5, 2)), [{}, {}, {"start": 0, "stop": 5, "step": 2}], 3),
    (
        (slice(0, 5, 2), Ellipsis, slice(-10, -20, -1)),
        [
            {"start": 0, "stop": 5, "step": 2},
            {},
            {},
            {},
            {"start": -10, "stop": -20, "step": -1},
        ],
        5,
    ),
    (
        (slice(0, 5, 2), Ellipsis, slice(-10, -20, -1)),
        [{"start": 0, "stop": 5, "step": 2}, {"start": -10, "stop": -20, "step": -1}],
        2,
    ),
]


@pytest.mark.parametrize("as_tuple, as_json, ndim", slice_json_cases)
def test_ndslice_to_json(as_tuple, as_json, ndim):
    assert NDSlice(*as_tuple).to_json(ndim=ndim) == as_json


def test_errors_in_slices():
    # Multiple Ellipsis are not allowed
    with pytest.raises(ValueError):
        NDSlice(Ellipsis, slice(1), Ellipsis)

    # Unspecified number of dimensions when converting to JSON
    with pytest.raises(ValueError):
        NDSlice(Ellipsis, slice(1)).to_json()

    # Dimension mismatch when converting to JSON
    with pytest.raises(ValueError):
        NDSlice(slice(1), slice(2), slice(3)).to_json(ndim=2)
    with pytest.raises(ValueError):
        NDSlice(slice(1), slice(2), Ellipsis).to_json(ndim=1)



import hashlib
import secrets
import uuid as uuid_module
import warnings
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Annotated, Any, Optional, Sequence, Set

from fastapi import (
    APIRouter,
    Depends,
    Form,
    Header,
    HTTPException,
    Query,
    Request,
    Response,
    Security,
    WebSocket,
)
from fastapi.security import (
    OAuth2PasswordBearer,
    OAuth2PasswordRequestForm,
    SecurityScopes,
)
from fastapi.security.api_key import APIKeyCookie, APIKeyHeader, APIKeyQuery
from fastapi.security.utils import get_authorization_scheme_param
from fastapi.templating import Jinja2Templates
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import selectinload
from sqlalchemy.sql import func
from starlette.status import (
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_404_NOT_FOUND,
    HTTP_409_CONFLICT,
)

from tiled.access_control.scopes import NO_SCOPES, PUBLIC_SCOPES, USER_SCOPES

# To hide third-party warning
# .../jose/backends/cryptography_backend.py:18: CryptographyDeprecationWarning:
#     int_from_bytes is deprecated, use int.from_bytes instead
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from jose import ExpiredSignatureError, JWTError, jwt

from pydantic import BaseModel

from ..authn_database import orm
from ..authn_database.connection_pool import get_database_session
from ..authn_database.core import (
    create_service,
    create_user,
    latest_principal_activity,
    lookup_valid_api_key,
    lookup_valid_pending_session_by_device_code,
    lookup_valid_pending_session_by_user_code,
    lookup_valid_session,
)
from ..utils import SHARE_TILED_PATH, SingleUserPrincipal
from . import schemas
from .core import DEFAULT_PAGE_SIZE, MAX_PAGE_SIZE, json_or_msgpack
from .protocols import ExternalAuthenticator, InternalAuthenticator, UserSessionState
from .settings import Settings, get_settings
from .utils import API_KEY_COOKIE_NAME, get_base_url

ALGORITHM = "HS256"

# Max API keys and Sessions allowed to Principal.
# This is here for at least two reasons:
# 1. Ensure that the routes which list API keys and sessions, which are
#    not paginated, returns in a reasonable time.
# 2. Avoid unintentional or intentional abuse.
API_KEY_LIMIT = 100
SESSION_LIMIT = 200

DEVICE_CODE_MAX_AGE = timedelta(minutes=15)
DEVICE_CODE_POLLING_INTERVAL = 5  # seconds


def utcnow():
    "UTC now with second resolution"
    return datetime.now(timezone.utc).replace(microsecond=0)


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    username: Optional[str] = None


class StrictAPIKeyHeader(APIKeyHeader):
    # TODO: remove custom subclass https://github.com/bluesky/tiled/issues/921
    """
    APIKeyHeader does not enforce that the scheme matches the expected, potentially
    leading to the case where Bearer tokens are treated as API keys.
    Additionally strips the scheme, as expected by current handling.
    """

    async def __call__(self, request: Request) -> Optional[str]:
        api_key: Optional[str] = request.headers.get(self.model.name)
        scheme, param = get_authorization_scheme_param(api_key)
        if not scheme or scheme.lower() == "bearer":
            return self.check_api_key(None, self.auto_error)
        if scheme.lower() != self.scheme_name.lower():
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=(
                    "Authorization header must include the authorization type "
                    "followed by a space and then the secret, as in "
                    "'Bearer SECRET' or 'Apikey SECRET'. "
                ),
            )
        return self.check_api_key(param, self.auto_error)


# The tokenUrl below is patched at app startup when we know it.
oauth2_scheme = OAuth2PasswordBearer(tokenUrl="PLACEHOLDER", auto_error=False)


def create_access_token(data, secret_key, expires_delta):
    to_encode = data.copy()
    expire = utcnow() + expires_delta
    to_encode.update({"exp": expire, "type": "access"})
    encoded_jwt = jwt.encode(to_encode, secret_key, algorithm=ALGORITHM)
    return encoded_jwt


def create_refresh_token(session_id, secret_key, expires_delta):
    expire = utcnow() + expires_delta
    to_encode = {
        "type": "refresh",
        "sid": session_id,
        "exp": expire,
    }
    encoded_jwt = jwt.encode(to_encode, secret_key, algorithm=ALGORITHM)
    return encoded_jwt


def decode_token(token, secret_keys):
    credentials_exception = HTTPException(
        status_code=HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    # The first key in settings.secret_keys is used for *encoding*.
    # All keys are tried for *decoding* until one works or they all
    # fail. They supports key rotation.
    for secret_key in secret_keys:
        try:
            payload = jwt.decode(token, secret_key, algorithms=[ALGORITHM])
            break
        except ExpiredSignatureError:
            # Do not let this be caught below with the other JWTError types.
            raise
        except JWTError:
            # Try the next key in the key rotation.
            continue
    else:
        raise credentials_exception
    return payload


async def get_api_key(
    api_key_query: Optional[str] = Depends(
        APIKeyQuery(name="api_key", auto_error=False)
    ),
    api_key_header: Optional[str] = Depends(
        StrictAPIKeyHeader(
            name="Authorization",
            description="Prefix value with 'Apikey ' as in, 'Apikey SECRET'",
            scheme_name="Apikey",
            auto_error=False,
        )
    ),
    api_key_cookie: Optional[str] = Depends(
        APIKeyCookie(name=API_KEY_COOKIE_NAME, auto_error=False)
    ),
) -> Optional[str]:
    for api_key in [api_key_query, api_key_header, api_key_cookie]:
        if api_key is not None:
            return api_key
    return None


def headers_for_401(request: Request, security_scopes: SecurityScopes):
    # call directly from methods, rather than as a dependency, to avoid calling
    # when not needed.
    if security_scopes.scopes:
        authenticate_value = f'Bearer scope="{security_scopes.scope_str}"'
    else:
        authenticate_value = "Bearer"
    headers_for_401 = {
        "WWW-Authenticate": authenticate_value,
        "X-Tiled-Root": get_base_url(request),
    }
    return headers_for_401


async def get_decoded_access_token(
    request: Request,
    security_scopes: SecurityScopes,
    access_token: str = Depends(oauth2_scheme),
    settings: Settings = Depends(get_settings),
):
    if not access_token:
        return None
    try:
        payload = decode_token(access_token, settings.secret_keys)
    except ExpiredSignatureError:
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED,
            detail="Access token has expired. Refresh token.",
            headers=headers_for_401(request, security_scopes),
        )
    return payload


async def get_session_state(decoded_access_token=Depends(get_decoded_access_token)):
    if decoded_access_token:
        return decoded_access_token.get("state")


async def get_access_tags_from_api_key(
    api_key: str, authenticated: bool, db: Optional[AsyncSession]
) -> Optional[Set[str]]:
    if not authenticated:
        # Tiled is in a "single user" mode with only one API key.
        # In this mode, there is no meaningful access tag limit.
        return None
    # Tiled is in a multi-user configuration with authentication providers.
    # We store the hashed value of the API key secret.
    try:
        secret = bytes.fromhex(api_key)
    except Exception:
        # access tag limit cannot be enforced without key information
        return None
    api_key_orm = await lookup_valid_api_key(db, secret)
    if api_key_orm is None:
        # access tag limit cannot be enforced without key information
        return None
    else:
        if (access_tags := api_key_orm.access_tags) is not None:
            access_tags = set(access_tags)
        return access_tags


async def get_current_access_tags(
    request: Request,
    api_key: Optional[str] = Depends(get_api_key),
    db: Optional[AsyncSession] = Depends(get_database_session),
) -> Optional[Set[str]]:
    if api_key is not None:
        return await get_access_tags_from_api_key(
            api_key, request.app.state.authenticated, db
        )
    else:
        # Limits on access tags only available via API key auth
        return None


def get_api_key_websocket(
    authorization: Annotated[Optional[str], Header()] = None,
):
    if authorization is None:
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED,
            detail="An API key must be passed in the Authorization header",
        )
    scheme, api_key = get_authorization_scheme_param(authorization)
    if scheme.lower() != "apikey":
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Authorization header must be formatted like 'Apikey SECRET'",
        )
    return api_key


async def get_current_access_tags_websocket(
    websocket: WebSocket,
    api_key: Optional[str] = Depends(get_api_key_websocket),
    db: Optional[AsyncSession] = Depends(get_database_session),
) -> Optional[Set[str]]:
    if api_key is not None:
        return await get_access_tags_from_api_key(
            api_key, websocket.app.state.authenticated, db
        )
    else:
        # Limits on access tags only available via API key auth
        return None


async def move_api_key(request: Request, api_key: Optional[str] = Depends(get_api_key)):
    """
    Move API key from query parameter to cookie.

    When a URL with an API key in the query parameter is opened in a browser,
    the API key is set as a cookie so that subsequent requests from the browser
    are authenticated. (This approach was inspired by Jupyter notebook.)
    """
    if ("api_key" in request.query_params) and (
        request.cookies.get(API_KEY_COOKIE_NAME) != api_key
    ):
        request.state.cookies_to_set.append(
            {"key": API_KEY_COOKIE_NAME, "value": api_key}
        )


async def get_scopes_from_api_key(
    api_key: str, settings: Settings, authenticated: bool, db: Optional[AsyncSession]
) -> Sequence[str]:
    if not authenticated:
        # Tiled is in a "single user" mode with only one API key.
        return (
            USER_SCOPES
            if secrets.compare_digest(api_key, settings.single_user_api_key)
            else set()
        )
    # Tiled is in a multi-user configuration with authentication providers.
    # We store the hashed value of the API key secret.
    # By comparing hashes we protect against timing attacks.
    # By storing only the hash of the (high-entropy) secret
    # we reduce the value of that an attacker can extracted from a
    # stolen database backup.
    try:
        secret = bytes.fromhex(api_key)
    except Exception:
        return NO_SCOPES
    api_key_orm = await lookup_valid_api_key(db, secret)
    if api_key_orm is None:
        return NO_SCOPES
    else:
        principal = api_key_orm.principal
        principal_scopes = set().union(*[role.scopes for role in principal.roles])
        # This intersection addresses the case where the Principal has
        # lost a scope that they had when this key was created.
        scopes = set(api_key_orm.scopes).intersection(principal_scopes | {"inherit"})
        if "inherit" in scopes:
            # The scope "inherit" is a metascope that confers all the
            # scopes for the Principal associated with this API,
            # resolved at access time.
            scopes.update(principal_scopes)
        return scopes


async def get_current_scopes(
    request: Request,
    decoded_access_token: Optional[dict[str, Any]] = Depends(get_decoded_access_token),
    api_key: Optional[str] = Depends(get_api_key),
    settings: Settings = Depends(get_settings),
    db: Optional[AsyncSession] = Depends(get_database_session),
) -> set[str]:
    if api_key is not None:
        return await get_scopes_from_api_key(
            api_key, settings, request.app.state.authenticated, db
        )
    elif decoded_access_token is not None:
        return decoded_access_token["scp"]
    else:
        return PUBLIC_SCOPES if settings.allow_anonymous_access else NO_SCOPES


async def get_current_scopes_websocket(
    websocket: WebSocket,
    api_key: Optional[str] = Depends(get_api_key_websocket),
    settings: Settings = Depends(get_settings),
    db: Optional[AsyncSession] = Depends(get_database_session),
) -> set[str]:
    if api_key is not None:
        return await get_scopes_from_api_key(
            api_key, settings, websocket.app.state.authenticated, db
        )
    else:
        return PUBLIC_SCOPES if settings.allow_anonymous_access else NO_SCOPES


async def check_scopes(
    request: Request,
    security_scopes: SecurityScopes,
    scopes: set[str] = Depends(get_current_scopes),
) -> None:
    if not set(security_scopes.scopes).issubset(scopes):
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED,
            detail=(
                "Not enough permissions. "
                f"Requires scopes {security_scopes.scopes}. "
                f"Request had scopes {list(scopes)}"
            ),
            headers=headers_for_401(request, security_scopes),
        )


async def get_current_principal_from_api_key(
    api_key: str, authenticated: bool, db: AsyncSession, settings: Settings
):
    if authenticated:
        # Tiled is in a multi-user configuration with authentication providers.
        # We store the hashed value of the API key secret.
        # By comparing hashes we protect against timing attacks.
        # By storing only the hash of the (high-entropy) secret
        # we reduce the value of that an attacker can extracted from a
        # stolen database backup.
        try:
            secret = bytes.fromhex(api_key)
        except Exception:
            # Not valid hex, therefore not a valid API key
            return None

        api_key_orm = await lookup_valid_api_key(db, secret)
        if api_key_orm is not None:
            principal = api_key_orm.principal
            principal_scopes = set().union(*[role.scopes for role in principal.roles])
            # This intersection addresses the case where the Principal has
            # lost a scope that they had when this key was created.
            scopes = set(api_key_orm.scopes).intersection(
                principal_scopes | {"inherit"}
            )
            if "inherit" in scopes:
                # The scope "inherit" is a metascope that confers all the
                # scopes for the Principal associated with this API,
                # resolved at access time.
                scopes.update(principal_scopes)
            api_key_orm.latest_activity = utcnow()
            await db.commit()
            return principal
        else:
            return None
    else:
        # Tiled is in a "single user" mode with only one API key.
        if secrets.compare_digest(api_key, settings.single_user_api_key):
            # Valid single-user API key - return None to indicate valid single user
            return None
        else:
            # Invalid single-user API key - raise exception directly
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED, detail="Invalid API key"
            )


async def get_current_principal_websocket(
    websocket: WebSocket,
    api_key: str = Depends(get_api_key_websocket),
    settings: Settings = Depends(get_settings),
    db: Optional[AsyncSession] = Depends(get_database_session),
):
    principal = await get_current_principal_from_api_key(
        api_key, websocket.app.state.authenticated, db, settings
    )
    if principal is None and websocket.app.state.authenticated:
        raise HTTPException(status_code=HTTP_401_UNAUTHORIZED, detail="Invalid API key")
    return principal


async def get_current_principal(
    request: Request,
    security_scopes: SecurityScopes,
    decoded_access_token: str = Depends(get_decoded_access_token),
    api_key: str = Depends(get_api_key),
    settings: Settings = Depends(get_settings),
    db: Optional[AsyncSession] = Depends(get_database_session),
    _=Depends(move_api_key),
) -> Optional[schemas.Principal]:
    """
    Get current Principal from:
    - API key in 'api_key' query parameter
    - API key in header 'Authorization: Apikey ...'
    - API key in cookie 'tiled_api_key'
    - OAuth2 JWT access token in header 'Authorization: Bearer ...'

    If anonymous access is allowed, Principal will be `None`.
    If the server is configured with a "single-user API key", then
    the Principal will also be `None` - but is differentiated for
    logging with a SingleUserPrincipal sentinel
    """

    if api_key is not None:
        principal = await get_current_principal_from_api_key(
            api_key,
            request.app.state.authenticated,
            db,
            settings,
        )
        if principal is None and request.app.state.authenticated:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail="Invalid API key",
                headers=headers_for_401(request, security_scopes),
            )
    elif decoded_access_token is not None:
        principal = schemas.Principal(
            uuid=uuid_module.UUID(hex=decoded_access_token["sub"]),
            type=decoded_access_token["sub_typ"],
            identities=[
                schemas.Identity(id=identity["id"], provider=identity["idp"])
                for identity in decoded_access_token["ids"]
            ],
        )
    else:
        # No form of authentication is present.
        principal = None
    # This is used to pass the currently-authenticated principal into the logger.
    is_apikey_single_user = api_key is not None and not request.app.state.authenticated
    request.state.principal = (
        principal if not is_apikey_single_user else SingleUserPrincipal
    )
    return principal


async def create_pending_session(db: AsyncSession):
    device_code = secrets.token_bytes(32)
    hashed_device_code = hashlib.sha256(device_code).digest()
    for _ in range(3):
        user_code = secrets.token_hex(4).upper()  # 8 digit code
        pending_session = orm.PendingSession(
            user_code=user_code,
            hashed_device_code=hashed_device_code,
            expiration_time=utcnow() + DEVICE_CODE_MAX_AGE,
        )
        db.add(pending_session)
        try:
            await db.commit()
        except IntegrityError:
            # Since the user_code is short, we cannot completely dismiss the
            # possibility of a collission. Retry.
            continue
        break
    formatted_user_code = f"{user_code[:4]}-{user_code[4:]}"
    return {
        "user_code": formatted_user_code,
        "device_code": device_code.hex(),
    }


async def create_session(
    settings: Settings,
    db: AsyncSession,
    identity_provider,
    id,
    state: UserSessionState = None,
):
    # Have we seen this Identity before?
    identity = (
        await db.execute(
            select(orm.Identity)
            .options(selectinload(orm.Identity.principal))
            .filter(orm.Identity.id == id)
            .filter(orm.Identity.provider == identity_provider)
        )
    ).scalar()
    now = utcnow()
    if identity is None:
        # We have not. Make a new Principal and link this new Identity to it.
        # TODO Confirm that the user intends to create a new Principal here.
        # Give them the opportunity to link an existing Principal instead.
        principal = await create_user(db, identity_provider, id)
        (new_identity,) = principal.identities
        new_identity.latest_login = now
    else:
        identity.latest_login = now
        principal = identity.principal
    session_count = (
        await db.execute(
            select(func.count())
            .select_from(orm.Session)
            .join(orm.Principal)
            .filter(orm.Principal.id == principal.id)
        )
    ).scalar()
    if session_count >= SESSION_LIMIT:
        raise HTTPException(
            400,
            f"This Principal already has {session_count} sessions which is greater "
            f"than or equal to the maximum number allowed, {SESSION_LIMIT}. "
            "Some Sessions must be closed before creating new ones.",
        )
    session = orm.Session(
        principal_id=principal.id,
        expiration_time=utcnow() + settings.session_max_age,
        state=state or {},
    )
    db.add(session)
    await db.commit()
    # Reload to select Principal and Identities.
    fully_loaded_session = (
        await db.execute(
            select(orm.Session)
            .options(
                selectinload(orm.Session.principal).selectinload(
                    orm.Principal.identities
                ),
            )
            .filter(orm.Session.id == session.id)
        )
    ).scalar()
    return fully_loaded_session


async def create_tokens_from_session(
    settings: Settings, db: AsyncSession, session, provider
):
    # Provide enough information in the access token to reconstruct Principal
    # and its Identities sufficient for access policy enforcement without a
    # database hit.
    principal = session.principal
    data = {
        "sub": principal.uuid.hex,
        "sub_typ": principal.type,  # Why is this str and not Enum?
        "scp": list(set().union(*[role.scopes for role in principal.roles])),
        "state": session.state,
        "ids": [
            {"id": identity.id, "idp": identity.provider}
            for identity in principal.identities
        ],
    }
    access_token = create_access_token(
        data=data,
        expires_delta=settings.access_token_max_age,
        secret_key=settings.secret_keys[0],  # Use the *first* secret key to encode.
    )
    refresh_token = create_refresh_token(
        session_id=session.uuid.hex,
        expires_delta=settings.refresh_token_max_age,
        secret_key=settings.secret_keys[0],  # Use the *first* secret key to encode.
    )
    # Include the identity. This is not stored as part of the session.
    # Once you are logged in, it does not matter *how* you logged in.
    # But in order to enable UIs to display a sensible username we provide
    # this information alongside the tokens only when the session is first created.
    identity = (
        await db.execute(
            select(orm.Identity)
            .filter(orm.Identity.principal == principal)
            .filter(orm.Identity.provider == provider)
        )
    ).scalar()
    return {
        "access_token": access_token,
        "expires_in": settings.access_token_max_age.total_seconds(),
        "refresh_token": refresh_token,
        "refresh_token_expires_in": settings.refresh_token_max_age.total_seconds(),
        "token_type": "bearer",
        "identity": {"id": identity.id, "provider": provider},
        "principal": principal.uuid.hex,
    }


def add_external_routes(
    router: APIRouter, provider: str, authenticator: ExternalAuthenticator
):
    if not SHARE_TILED_PATH:
        raise Exception(
            "Static assets could not be found and are required for "
            "setting up external OAuth authentication."
        )
    templates = Jinja2Templates(Path(SHARE_TILED_PATH, "templates"))

    "Build an auth_code route function for this Authenticator."

    @router.get(f"/provider/{provider}/code")
    async def auth_code_route(
        request: Request,
        settings: Settings = Depends(get_settings),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        request.state.endpoint = "auth"
        user_session_state: Optional[
            UserSessionState
        ] = await authenticator.authenticate(request)
        if not user_session_state:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED, detail="Authentication failure"
            )
        session = await create_session(
            settings,
            db,
            provider,
            user_session_state.user_name,
            user_session_state.state,
        )
        tokens = await create_tokens_from_session(settings, db, session, provider)
        return tokens

    "Build an /authorize route function for this Authenticator."

    @router.post(f"/provider/{provider}/authorize")
    async def device_code_authorize_route(
        request: Request,
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        request.state.endpoint = "auth"
        pending_session = await create_pending_session(db)
        verification_uri = f"{get_base_url(request)}/auth/provider/{provider}/token"
        authorization_uri = authenticator.authorization_endpoint.copy_with(
            params={
                "client_id": authenticator.client_id,
                "response_type": "code",
                "scope": "openid",
                "redirect_uri": f"{get_base_url(request)}/auth/provider/{provider}/device_code",
            }
        )
        return {
            "authorization_uri": str(
                authorization_uri
            ),  # URL that user should visit in browser
            "verification_uri": str(
                verification_uri
            ),  # URL that terminal client will poll
            "interval": DEVICE_CODE_POLLING_INTERVAL,  # suggested polling interval
            "device_code": pending_session["device_code"],
            "expires_in": DEVICE_CODE_MAX_AGE,  # seconds
            "user_code": pending_session["user_code"],
        }

    @router.get(f"/provider/{provider}/device_code")
    async def device_code_user_code_form_route(
        request: Request,
        code: str,
    ):
        action = (
            f"{get_base_url(request)}/auth/provider/{provider}/device_code?code={code}"
        )
        return templates.TemplateResponse(
            request,
            "device_code_form.html",
            {
                "code": code,
                "action": action,
            },
        )

    "Build an /authorize route function for this Authenticator."

    @router.post(f"/provider/{provider}/device_code")
    async def device_code_user_code_submit_route(
        request: Request,
        code: str = Form(),
        user_code: str = Form(),
        settings: Settings = Depends(get_settings),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        request.state.endpoint = "auth"
        action = (
            f"{get_base_url(request)}/auth/provider/{provider}/device_code?code={code}"
        )
        normalized_user_code = user_code.upper().replace("-", "").strip()
        pending_session = await lookup_valid_pending_session_by_user_code(
            db, normalized_user_code
        )
        if pending_session is None:
            message = "Invalid user code. It may have been mistyped, or the pending request may have expired."
            return templates.TemplateResponse(
                request,
                "device_code_form.html",
                {
                    "code": code,
                    "action": action,
                    "message": message,
                },
                status_code=HTTP_401_UNAUTHORIZED,
            )
        user_session_state: Optional[
            UserSessionState
        ] = await authenticator.authenticate(request)
        if not user_session_state:
            return templates.TemplateResponse(
                request,
                "device_code_failure.html",
                {
                    "message": (
                        "User code was correct but authentication with third party failed. "
                        "Ask administrator to see logs for details."
                    ),
                },
                status_code=HTTP_401_UNAUTHORIZED,
            )
        session = await create_session(
            settings,
            db,
            provider,
            user_session_state.user_name,
            user_session_state.state,
        )
        pending_session.session_id = session.id
        db.add(pending_session)
        await db.commit()
        return templates.TemplateResponse(
            request,
            "device_code_success.html",
            {
                "interval": DEVICE_CODE_POLLING_INTERVAL,
            },
        )

    "Build an /authorize route function for this Authenticator."

    @router.post(f"/provider/{provider}/token")
    async def device_code_token_route(
        request: Request,
        body: schemas.DeviceCode,
        settings: Settings = Depends(get_settings),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        request.state.endpoint = "auth"
        device_code_hex = body.device_code
        try:
            device_code = bytes.fromhex(device_code_hex)
        except Exception:
            # Not valid hex, therefore not a valid device_code
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED, detail="Invalid device code"
            )
        pending_session = await lookup_valid_pending_session_by_device_code(
            db, device_code
        )
        if pending_session is None:
            raise HTTPException(
                404,
                detail="No such device_code. The pending request may have expired.",
            )
        if pending_session.session_id is None:
            raise HTTPException(
                HTTP_400_BAD_REQUEST, {"error": "authorization_pending"}
            )
        session = pending_session.session
        # The pending session can only be used once.
        await db.delete(pending_session)
        await db.commit()
        tokens = await create_tokens_from_session(settings, db, session, provider)
        return tokens


def add_internal_routes(
    router: APIRouter, provider: str, authenticator: InternalAuthenticator
):
    "Register a handle_credentials route function for this Authenticator."

    @router.post(f"/provider/{provider}/token")
    async def handle_credentials_route(
        request: Request,
        form_data: OAuth2PasswordRequestForm = Depends(),
        settings: Settings = Depends(get_settings),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        request.state.endpoint = "auth"
        user_session_state: Optional[
            UserSessionState
        ] = await authenticator.authenticate(
            username=form_data.username, password=form_data.password
        )
        if not user_session_state or not user_session_state.user_name:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail="Incorrect username or password",
                headers={"WWW-Authenticate": "Bearer"},
            )
        session = await create_session(
            settings,
            db,
            provider,
            user_session_state.user_name,
            state=user_session_state.state,
        )
        tokens = await create_tokens_from_session(settings, db, session, provider)
        return tokens


async def generate_apikey(db: AsyncSession, principal, apikey_params, request):
    if apikey_params.scopes is None:
        scopes = ["inherit"]
    else:
        scopes = apikey_params.scopes
    principal_scopes = set().union(*[role.scopes for role in principal.roles])
    if not set(scopes).issubset(principal_scopes | {"inherit"}):
        raise HTTPException(
            403,
            (
                f"Requested scopes {apikey_params.scopes} must be a subset of the "
                f"principal's scopes {list(principal_scopes)}."
            ),
        )
    admin_scopes = ["admin:apikeys"]
    if (access_tags := apikey_params.access_tags) is not None:
        if all(scope in scopes for scope in admin_scopes):
            raise HTTPException(
                403,
                (
                    f"Requested scopes {scopes} contain scopes {admin_scopes}, "
                    f"which cannot be combined with access tag restrictions."
                ),
            )
    if apikey_params.expires_in is not None:
        expiration_time = utcnow() + timedelta(seconds=apikey_params.expires_in)
    else:
        expiration_time = None
    # The standard 32 byes of entropy,
    # plus 4 more for extra safety since we store the first eight HEX chars.
    secret = secrets.token_bytes(4 + 32)
    hashed_secret = hashlib.sha256(secret).digest()
    keys_count = (
        await db.execute(
            select(func.count())
            .select_from(orm.APIKey)
            .join(orm.Principal)
            .filter(orm.Principal.id == principal.id)
        )
    ).scalar()
    if keys_count >= API_KEY_LIMIT:
        raise HTTPException(
            400,
            f"This Principal already has {keys_count} API keys which is greater "
            f"than or equal to the maximum number allowed, {API_KEY_LIMIT}. "
            "Some API keys must be deleted before creating new ones.",
        )
    new_key = orm.APIKey(
        principal_id=principal.id,
        expiration_time=expiration_time,
        note=apikey_params.note,
        scopes=scopes,
        access_tags=access_tags,
        first_eight=secret.hex()[:8],
        hashed_secret=hashed_secret,
    )
    db.add(new_key)
    await db.commit()
    # db.refresh(new_key)
    return json_or_msgpack(
        request,
        schemas.APIKeyWithSecret.from_orm(new_key, secret=secret.hex()).model_dump(),
    )


def authentication_router() -> APIRouter:
    router = APIRouter()

    @router.get(
        "/principal",
        response_model=schemas.Principal,
    )
    async def principal_list(
        request: Request,
        offset: Optional[int] = Query(0, alias="page[offset]", ge=0),
        limit: Optional[int] = Query(
            DEFAULT_PAGE_SIZE, alias="page[limit]", ge=0, le=MAX_PAGE_SIZE
        ),
        principal: Optional[schemas.Principal] = Depends(get_current_principal),
        _=Security(check_scopes, scopes=["read:principals"]),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        "List Principals (users and services)."
        request.state.endpoint = "auth"
        principal_orms = (
            (
                await db.execute(
                    select(orm.Principal)
                    .offset(offset)
                    .limit(limit)
                    .options(
                        selectinload(orm.Principal.identities),
                        selectinload(orm.Principal.roles),
                        selectinload(orm.Principal.api_keys),
                        selectinload(orm.Principal.sessions),
                    )
                )
            )
            .unique()
            .all()
        )
        principals = []
        for (principal_orm,) in principal_orms:
            latest_activity = await latest_principal_activity(db, principal_orm)
            principal = schemas.Principal.from_orm(
                principal_orm, latest_activity
            ).model_dump()
            principals.append(principal)
        return json_or_msgpack(request, principals)

    @router.post(
        "/principal",
        response_model=schemas.Principal,
    )
    async def create_service_principal(
        request: Request,
        principal: Optional[schemas.Principal] = Depends(get_current_principal),
        _=Security(check_scopes, scopes=["write:principals"]),
        db: Optional[AsyncSession] = Depends(get_database_session),
        role: str = Query(...),
    ):
        "Create a principal for a service account."

        principal_orm = await create_service(db, role)

        # Reload to select Principal and Identities.
        fully_loaded_principal_orm = (
            await db.execute(
                select(orm.Principal)
                .options(
                    selectinload(orm.Principal.identities),
                    selectinload(orm.Principal.roles),
                    selectinload(orm.Principal.api_keys),
                    selectinload(orm.Principal.sessions),
                )
                .filter(orm.Principal.id == principal_orm.id)
            )
        ).scalar()

        principal = schemas.Principal.from_orm(fully_loaded_principal_orm).model_dump()
        request.state.endpoint = "auth"

        return json_or_msgpack(request, principal)

    @router.get(
        "/principal/{uuid}",
        response_model=schemas.Principal,
    )
    async def principal(
        request: Request,
        uuid: uuid_module.UUID,
        _=Security(check_scopes, scopes=["read:principals"]),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        "Get information about one Principal (user or service)."
        request.state.endpoint = "auth"
        principal_orm = (
            await db.execute(
                select(orm.Principal)
                .filter(orm.Principal.uuid == uuid)
                .options(
                    selectinload(orm.Principal.identities),
                    selectinload(orm.Principal.roles),
                    selectinload(orm.Principal.api_keys),
                    selectinload(orm.Principal.sessions),
                )
            )
        ).scalar()
        if principal_orm is None:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND, detail=f"No such Principal {uuid}"
            )
        latest_activity = await latest_principal_activity(db, principal_orm)

        return json_or_msgpack(
            request,
            schemas.Principal.from_orm(principal_orm, latest_activity).model_dump(),
        )

    @router.delete(
        "/principal/{uuid}/apikey",
        response_model=schemas.Principal,
    )
    async def revoke_apikey_for_principal(
        request: Request,
        uuid: uuid_module.UUID,
        first_eight: str,
        _=Security(check_scopes, scopes=["admin:apikeys"]),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        "Allow Tiled Admins to delete any user's apikeys e.g."
        request.state.endpoint = "auth"
        api_key_orm = (
            await db.execute(
                select(orm.APIKey).filter(orm.APIKey.first_eight == first_eight[:8])
            )
        ).scalar()
        if (api_key_orm is None) or (api_key_orm.principal.uuid != uuid):
            raise HTTPException(
                404,
                f"The principal {uuid} has no such API key.",
            )
        await db.delete(api_key_orm)
        await db.commit()

        return Response(status_code=HTTP_204_NO_CONTENT)

    @router.post(
        "/principal/{uuid}/apikey",
        response_model=schemas.APIKeyWithSecret,
    )
    async def apikey_for_principal(
        request: Request,
        uuid: uuid_module.UUID,
        apikey_params: schemas.APIKeyRequestParams,
        principal: Optional[schemas.Principal] = Depends(get_current_principal),
        _=Security(check_scopes, scopes=["admin:apikeys"]),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        "Generate an API key for a Principal."
        request.state.endpoint = "auth"
        principal = (
            await db.execute(select(orm.Principal).filter(orm.Principal.uuid == uuid))
        ).scalar()
        if principal is None:
            raise HTTPException(
                404, f"Principal {uuid} does not exist or insufficient permissions."
            )
        return await generate_apikey(db, principal, apikey_params, request)

    @router.post("/session/refresh", response_model=schemas.AccessAndRefreshTokens)
    async def refresh_session(
        request: Request,
        refresh_token: schemas.RefreshToken,
        settings: Settings = Depends(get_settings),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        "Obtain a new access token and refresh token."
        request.state.endpoint = "auth"
        new_tokens = await slide_session(refresh_token.refresh_token, settings, db)
        return new_tokens

    @router.post("/session/revoke")
    async def revoke_session(
        request: Request,
        refresh_token: schemas.RefreshToken,
        settings: Settings = Depends(get_settings),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        "Mark a Session as revoked so it cannot be refreshed again."
        request.state.endpoint = "auth"
        payload = decode_token(refresh_token.refresh_token, settings.secret_keys)
        session_id = payload["sid"]
        # Find this session in the database.
        session = await lookup_valid_session(db, session_id)
        if session is None:
            raise HTTPException(HTTP_409_CONFLICT, detail=f"No session {session_id}")
        session.revoked = True
        db.add(session)
        await db.commit()
        return Response(status_code=HTTP_204_NO_CONTENT)

    @router.delete("/session/revoke/{session_id}")
    async def revoke_session_by_id(
        session_id: str,  # from path parameter
        request: Request,
        principal: Optional[schemas.Principal] = Depends(get_current_principal),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        "Mark a Session as revoked so it cannot be refreshed again."
        request.state.endpoint = "auth"
        # Find this session in the database.
        session = await lookup_valid_session(db, session_id)
        if session is None:
            raise HTTPException(404, detail=f"No session {session_id}")
        if principal.uuid != session.principal.uuid:
            # TODO Add a scope for doing this for other users.
            raise HTTPException(
                HTTP_404_NOT_FOUND,
                detail="Sessions does not exist or requester has insufficient permissions",
            )
        session.revoked = True
        db.add(session)
        await db.commit()
        return Response(status_code=HTTP_204_NO_CONTENT)

    async def slide_session(refresh_token, settings, db):
        try:
            payload = decode_token(refresh_token, settings.secret_keys)
        except ExpiredSignatureError:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail="Session has expired. Please re-authenticate.",
            )
        # Find this session in the database.
        session = await lookup_valid_session(db, payload["sid"])
        now = utcnow()
        # This token is *signed* so we know that the information came from us.
        # If the Session is forgotten or revoked or expired, do not allow refresh.
        if (
            (session is None)
            or session.revoked
            or (session.expiration_time.replace(tzinfo=timezone.utc) < now)
        ):
            # Do not leak (to a potential attacker) whether this has been *revoked*
            # specifically. Give the same error as if it had expired.
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail="Session has expired. Please re-authenticate.",
            )
        # Update Session info.
        session.time_last_refreshed = now
        # This increments in a way that avoids a race condition.
        session.refresh_count = orm.Session.refresh_count + 1
        # Update the database.
        db.add(session)
        await db.commit()
        # Provide enough information in the access token to reconstruct Principal
        # and its Identities sufficient for access policy enforcement without a
        # database hit.
        data = {
            "sub": session.principal.uuid.hex,
            "sub_typ": session.principal.type,  # Why is this str and not Enum?
            "scp": list(
                set().union(*[role.scopes for role in session.principal.roles])
            ),
            "state": session.state,
            "ids": [
                {"id": identity.id, "idp": identity.provider}
                for identity in session.principal.identities
            ],
        }
        access_token = create_access_token(
            data=data,
            expires_delta=settings.access_token_max_age,
            secret_key=settings.secret_keys[0],  # Use the *first* secret key to encode.
        )
        new_refresh_token = create_refresh_token(
            session_id=payload["sid"],
            expires_delta=settings.refresh_token_max_age,
            secret_key=settings.secret_keys[0],  # Use the *first* secret key to encode.
        )
        return {
            "access_token": access_token,
            "expires_in": settings.access_token_max_age.total_seconds(),
            "refresh_token": new_refresh_token,
            "refresh_token_expires_in": settings.refresh_token_max_age.total_seconds(),
            "token_type": "bearer",
        }

    @router.post(
        "/apikey",
        response_model=schemas.APIKeyWithSecret,
    )
    async def new_apikey(
        request: Request,
        apikey_params: schemas.APIKeyRequestParams,
        principal: Optional[schemas.Principal] = Depends(get_current_principal),
        _=Security(check_scopes, scopes=["apikeys"]),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        """
        Generate an API for the currently-authenticated user or service.
        """
        # TODO Permit filtering the fields of the response.
        request.state.endpoint = "auth"
        if principal is None:
            return None
        # The principal from get_current_principal tells us everything that the
        # access_token carries around, but the database knows more than that.
        principal_orm = (
            await db.execute(
                select(orm.Principal).filter(orm.Principal.uuid == principal.uuid)
            )
        ).scalar()
        apikey = await generate_apikey(db, principal_orm, apikey_params, request)
        return apikey

    @router.get("/apikey", response_model=schemas.APIKey)
    async def current_apikey_info(
        request: Request,
        api_key: str = Depends(get_api_key),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        """
        Give info about the API key used to authentication the current request.

        This provides a way to look up the API uuid, given the API secret.
        """
        # TODO Permit filtering the fields of the response.
        request.state.endpoint = "auth"
        if api_key is None:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail="No API key was provided with this request.",
            )
        try:
            secret = bytes.fromhex(api_key)
        except Exception:
            # Not valid hex, therefore not a valid API key
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED, detail="Invalid API key"
            )
        api_key_orm = await lookup_valid_api_key(db, secret)
        if api_key_orm is None:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED, detail="Invalid API key"
            )
        return json_or_msgpack(
            request, schemas.APIKey.from_orm(api_key_orm).model_dump()
        )

    @router.delete("/apikey")
    async def revoke_apikey(
        request: Request,
        first_eight: str,
        principal: Optional[schemas.Principal] = Depends(get_current_principal),
        _=Security(check_scopes, scopes=["apikeys"]),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        """
        Revoke an API belonging to the currently-authenticated user or service."""
        # TODO Permit filtering the fields of the response.
        request.state.endpoint = "auth"
        if principal is None:
            return None
        api_key_orm = (
            await db.execute(
                select(orm.APIKey).filter(orm.APIKey.first_eight == first_eight[:8])
            )
        ).scalar()
        if (api_key_orm is None) or (api_key_orm.principal.uuid != principal.uuid):
            raise HTTPException(
                404,
                f"The currently-authenticated {principal.type} has no such API key.",
            )
        await db.delete(api_key_orm)
        await db.commit()
        return Response(status_code=HTTP_204_NO_CONTENT)

    @router.get(
        "/whoami",
        response_model=schemas.Principal,
    )
    async def whoami(
        request: Request,
        principal: Optional[schemas.Principal] = Depends(get_current_principal),
        db: Optional[AsyncSession] = Depends(get_database_session),
    ):
        # TODO Permit filtering the fields of the response.
        request.state.endpoint = "auth"
        if principal is None:
            return json_or_msgpack(request, None)
        # The principal from get_current_principal tells us everything that the
        # access_token carries around, but the database knows more than that.
        principal_orm = (
            await db.execute(
                select(orm.Principal)
                .options(
                    selectinload(orm.Principal.identities),
                    selectinload(orm.Principal.roles),
                    selectinload(orm.Principal.api_keys),
                    selectinload(orm.Principal.sessions),
                )
                .filter(orm.Principal.uuid == principal.uuid)
            )
        ).scalar()
        if principal_orm is None:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED, detail="Principal no longer exists."
            )
        latest_activity = await latest_principal_activity(db, principal_orm)
        return json_or_msgpack(
            request,
            schemas.Principal.from_orm(principal_orm, latest_activity).model_dump(),
        )

    @router.post("/logout", include_in_schema=False, deprecated=True)
    async def logout(
        request: Request,
        response: Response,
    ):
        "Deprecated. See revoke_session: POST /session/revoke."
        request.state.endpoint = "auth"
        response.delete_cookie(API_KEY_COOKIE_NAME)
        return {}

    return router



from collections import Counter
from typing import Callable

from .structures.core import Spec, StructureFamily


class ValidationRegistry:
    """
    Register validation functions for specs
    """

    def __init__(self):
        self._lookup = {}

    def register(self, spec: Spec, func: Callable):
        self._lookup[spec] = func

    def dispatch(self, spec: Spec):
        try:
            return self._lookup[spec]
        except KeyError:
            pass
        raise ValueError(f"No dispatch for spec {spec}")

    def __call__(self, spec: Spec):
        return self.dispatch(spec)

    def __contains__(self, spec: Spec):
        return spec in self._lookup


class ValidationError(Exception):
    pass


async def validate_composite(spec, metadata, entry, structure_family, structure):
    """Spec validator for Composite containers

    Imposes rules on the contents of the entry (parent container):

    1. The spec can be assigned only to containers.
    2. No nested containers are allowed.
    3. The column names of all table structures must be unique and not
       conflict with the names of any other contents.
    """

    if entry is None:
        # Creating a new node (container)
        if structure_family != StructureFamily.container:
            raise ValidationError(
                f"Composite spec can be assigned only to containers, not to"
                f" {structure_family}."
            )
    else:
        # Updating an existing node (container)
        if entry.structure_family != StructureFamily.container:
            raise ValidationError(
                f"Composite spec can be assigned only to containers, not to"
                f" {entry.structure_family}."
            )

        flat_name_space = []
        for key, item in await entry.items_range(offset=0, limit=None):
            flat_name_space.append(key)

            if item.structure_family == StructureFamily.table:
                flat_name_space.extend(item.structure().columns)

            elif item.structure_family == StructureFamily.container:
                raise ValidationError(
                    "Nested containers are not allowed in a composite container."
                )

        counts = Counter(flat_name_space)
        if repeats := [item for item, count in counts.items() if count > 1]:
            raise ValidationError(
                "Names of table columns and other items in a composite container"
                " must be unique and not conflict with each other. "
                f"Found conflicting names: {repeats}."
            )


# Global instance of the validation registry
default_validation_registry = ValidationRegistry()
default_validation_registry.register(Spec("composite", None), validate_composite)



from typing import List, Optional

import typer

from ._utils import get_context

api_key_app = typer.Typer(no_args_is_help=True)


@api_key_app.command("create")
def create_api_key(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    expires_in: Optional[str] = typer.Option(
        None,
        help=(
            "Number of seconds until API key expires, given as integer seconds "
            "or a string like: '3y' (years), '3d' (days), '5m' (minutes), '1h' "
            "(hours), '30s' (seconds). If None, it will never expire or it will "
            "have the maximum lifetime allowed by the server. "
        ),
    ),
    scopes: Optional[List[str]] = typer.Option(
        None,
        help=(
            "Restrict the access available to this API key by listing scopes. "
            "By default, it will inherit the scopes of its owner."
        ),
    ),
    access_tags: Optional[List[str]] = typer.Option(
        None,
        help=(
            "Restrict the access available to the API key by listing specific tags. "
            "By default, it will have no limits on access tags."
        ),
    ),
    note: Optional[str] = typer.Option(None, help="Add a note to label this API key."),
    no_verify: bool = typer.Option(False, "--no-verify", help="Skip SSL verification."),
):
    context = get_context(profile)
    if not scopes:
        # This is how typer interprets unspecified scopes.
        # Replace with None to get default scopes.
        scopes = None
    if expires_in and expires_in.isdigit():
        expires_in = int(expires_in)
    info = context.create_api_key(
        scopes=scopes, access_tags=access_tags, expires_in=expires_in, note=note
    )
    # TODO Print other info to the stderr?
    typer.echo(info["secret"])


@api_key_app.command("list")
def list_api_keys(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
):
    context = get_context(profile)
    info = context.whoami()
    if not info["api_keys"]:
        typer.echo("No API keys found", err=True)
        return
    max_note_len = max(len(api_key["note"] or "") for api_key in info["api_keys"])
    if (starting_notes_pad := max_note_len) < len("Note"):
        starting_notes_pad += 4 - max_note_len
    max_scopes_len = max(
        sum(len(scope) for scope in api_key["scopes"]) + len(api_key["scopes"]) - 1
        for api_key in info["api_keys"]
    )
    if (starting_scopes_pad := max_scopes_len) < len("Scopes"):
        starting_scopes_pad += 6 - max_scopes_len
    COLUMNS = (
        f"First 8   Expires at (UTC)     "
        f"Latest activity      Note{' ' * (max_note_len - 4)}    "
        f"Scopes{' ' * (max_scopes_len - 6)}    Access tags"
    )
    typer.echo(COLUMNS)
    for api_key in info["api_keys"]:
        note_padding = 4 + starting_notes_pad - len(api_key["note"] or "")
        # the '1' subtraction works in all cases because the amount of spaces
        #   in a sentence is count(words) - 1, and also because we otherwise
        #   print a single 'dash' for an empty list
        scopes_padding = (
            4
            + starting_scopes_pad
            - sum(len(scope) for scope in api_key["scopes"])
            + len(api_key["scopes"])
            - 1
        )
        if api_key["expiration_time"] is None:
            expiration_time = "-"
        else:
            expiration_time = (
                api_key["expiration_time"]
                .replace(microsecond=0, tzinfo=None)
                .isoformat()
            )
        if api_key["latest_activity"] is None:
            latest_activity = "-"
        else:
            latest_activity = (
                api_key["latest_activity"]
                .replace(microsecond=0, tzinfo=None)
                .isoformat()
            )
        access_tags = (
            " ".join([tag.replace(" ", "\\ ") for tag in api_key["access_tags"]])
            if api_key["access_tags"] is not None
            else "-"
        )
        typer.echo(
            (
                f"{api_key['first_eight']:10}"
                f"{expiration_time:21}"
                f"{latest_activity:21}"
                f"{(api_key['note'] or '')}{' ' * note_padding}"
                f"{' '.join(api_key['scopes']) or '-'}{' ' * scopes_padding}"
                f"{access_tags}"
            )
        )


@api_key_app.command("revoke")
def revoke_api_key(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    first_eight: str = typer.Argument(
        ..., help="First eight characters of API key (or the whole key)"
    ),
):
    context = get_context(profile)
    context.revoke_api_key(first_eight[:8])
